from __future__ import annotations

import importlib.metadata
from typing import Any

project = "Pigreads"
copyright = "2024, Desmond Kabus"
author = "Desmond Kabus"
version = release = importlib.metadata.version("pigreads")

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
]

source_suffix = [".rst", ".md"]
exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]

html_theme = "furo"

html_theme_options: dict[str, Any] = {
    "source_repository": "https://gitlab.com/pigreads/pigreads",
    "source_branch": "main",
    "source_directory": "docs/",
}

myst_enable_extensions = [
    "colon_fence",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
}

nitpick_ignore = [
    ("py:class", "numpy.int32"),  # TODO: remove when possible
    ("py:class", "numpy.int64"),  # TODO: remove when possible
    ("py:class", "numpy.float32"),  # TODO: remove when possible
    ("py:class", "numpy.float64"),  # TODO: remove when possible
    ("py:class", "_io.StringIO"),
    ("py:class", "_io.BytesIO"),
]

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
    "no-value": True,
}

always_document_param_types = True
typehints_fully_qualified = False
typehints_defaults = "braces-after"
typehints_document_rtype = True



from __future__ import annotations

import importlib.metadata
from typing import Any

import numpy as np

import pigreads as pig


def test_version():
    assert importlib.metadata.version("pigreads") == pig.__version__


def test_deltas():
    z, y, x = np.meshgrid(
        [0], np.arange(0, 4, 0.2), np.linspace(0, 1, 5), indexing="ij"
    )
    dz, dy, dx = pig.deltas(z, y, x)
    assert dz == 1
    assert dy == 0.2
    assert dx == 0.25
    assert pig.delta(y, ax=-2) == dy


def test_ithildin():
    z, y, x = np.mgrid[0:1, 0:3, 0:3]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)
    models = pig.Models()
    models.add("marcotte2017dynamical")
    assert models[0]["variables"] == {"u": 0.0, "v": 0.0}
    states: np.ndarray[Any, Any] = models.resting_states(inhom, Nframes=2).astype(
        np.float32
    )
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=0)
    Nt, dt = 3, 0.01
    states[1] = pig.run(models, inhom, weights, states[0], Nt=Nt, dt=dt)
    log, variables = pig.to_ithildin(Nt * dt, dt, dz, dy, dx, models, states, inhom)
    del log["Start date"]
    del log["Simulation parameters"]["Ithildin library version"]
    del log["Simulation parameters"]["Serial number"]
    assert (
        str(log)
        == r"{'Ithildin log version': 2, 'Simulation parameters': {'Timestep dt': 0.01, 'Frame duration': 0.03, 'Number of frames to take': 2, 'Name of simulation series': 'pigreads'}, 'Geometry parameters': {'Number of dimensions': 3, 'Voxel size': [1.0, 1.0, 1.0], 'Domain size': [3, 3, 1]}, 'Model parameters': {'Model type': 'Smoothed Karma (Marcotte & Grigoriev 2017)', 'Class': 'marcotte2017dynamical', 'Citation': 'https://doi.org/10.1063/1.5003259\nhttps://doi.org/10.1063/1.4915143\nhttps://doi.org/10.1103/PhysRevLett.71.1103\nhttps://doi.org/10.1063/1.166024', 'Parameters': {'diffusivity_u': 4.0062, 'diffusivity_v': 0.20031, 'beta': 1.389, 'eps': 0.01, 'ustar': 1.5415}, 'Initial values': {'u': 0.0, 'v': 0.0}, 'Variable names': ['u', 'v'], 'Number of vars': 2}}"
    )
    assert " ".join(variables.keys()) == "u v inhom"
    assert variables["u"].shape == (2, 1, 3, 3)
    assert variables["v"].shape == (2, 1, 3, 3)



from __future__ import annotations

import numpy as np
import pytest

import pigreads as pig


def test_small():
    with pytest.raises(AssertionError):
        pig.normalise_vector(np.array([]))


def test_large():
    with pytest.raises(AssertionError):
        pig.normalise_vector(np.ones((1, 1, 1, 1, 1)))


def test_2():
    with pytest.raises(AssertionError):
        pig.normalise_vector(np.ones((2, 2, 2)))


def test_034():
    nvec = pig.normalise_vector([0, 3, 4])
    assert nvec.shape == (1, 1, 1, 3, 1), "5 dimensions: z, y, x, row, col"
    assert np.allclose(nvec.flatten(), [0, 0.6, 0.8])


def test_000():
    assert np.allclose(pig.normalise_vector([0, 0, 0]), [0, 0, 0])


def test_random():
    random = np.random.default_rng(seed=0)
    vec = random.uniform(size=(2, 4, 2, 3))
    nvec = pig.normalise_vector(vec)
    assert nvec.shape == (2, 4, 2, 3, 1), "5 dimensions: z, y, x, row, col"
    length = np.linalg.norm(nvec, axis=(-2))
    assert np.allclose(length, 1)



# Pigreads: Python-integrated GPU-enabled reaction diffusion solver
# Copyright (c) 2024 Desmond Kabus. All rights reserved.

"""
Pigreads Python module
----------------------

This Python module is the main interface to set up and run Pigreads simulations
solving the reaction-diffusion equations with OpenCL and NumPy:

.. math::

    \\partial_t \\underline{u}
    =
    \\underline{P} \\nabla \\cdot \\mathbf D \\nabla \\underline{u}
    +
    \\underline{r}(\\underline{u})

for :math:`\\underline{u}(t, \\mathbf x)`, :math:`t\\in[0, T]`, and
:math:`\\mathbf x\\in\\Omega\\subset\\mathbb R^3`, with initial conditions and
no-flux boundary conditions.

The following equations define a simpler example with only two variables,
:math:`\\underline{u} = (u, v)`, with no diffusion in :math:`v`, and
homogeneous and isotropic diffusion:

.. math::

    \\begin{aligned}
    \\partial_t u
    &=
    D \\nabla^2 u
    +
    r_u(u, v)
    \\\\
    \\partial_t v
    &=
    r_v(u, v)
    \\end{aligned}

Pigreads performs the most expensive calculations on graphics cards
using OpenCL, see :py:func:`run` and :py:func:`weights`. Input and output as well
as setting up and interacting with the simulation should be done in Python,
with the exception of adding source terms, so-called stimulus currents.
Pigreads uses the simplistic finite-differences method and forward-Euler time
stepping.

A Pigreads simulation is usually defined in the following steps:
First, define the geometry of the medium. In this example, we use a 2D plane
with 200 points in both x and y::

    import pigreads as pig
    import numpy as np

    R = 11.05
    z, y, x = np.mgrid[0:1, -R:R:200j, -R:R:200j]
    dz, dy, dx = pig.deltas(z, y, x)
    r = np.linalg.norm((x, y, z), axis=0)

Pigreads is optimised for three-dimensional space. For
lower-dimensional simulations, set the number of points in additional
dimensions to one, as done above for the z-dimension.

Calculations are performed at all points with periodic boundary conditions.
The integer field ``inhom`` defines which points are inside (1) the medium and outside (0)::

    inhom = np.ones_like(x, dtype=int)
    inhom[r > R] = 0

Values of inhom larger than zero can be used to select one of multiple models,
i.e., reaction terms :math:`\\underline{r}`. For an ``inhom`` value of 1,
``models[0]`` is used; and ``models[1]`` for a value of 2, etc. One or more of the
available models can be selected using an instance of the :py:class:`Models`
class::

    models = pig.Models()
    models.add("marcotte2017dynamical", diffusivity_u=1.0, diffusivity_v=0.05)

This class also has a function to create an array of the same shape as ``inhom``
in space but for a given number of frames in time. The first frame is filled
with the appropriate resting values for each model. Initial conditions can then
be set in the 0th frame::

    states = models.resting_states(inhom, Nframes=100)
    states[0, np.linalg.norm(((x + 8), y, z), axis=0) < 2, 0] = 1
    states[0, y < 0, 1] = 2

Note that states has five indices, in this order: time, z, y, x, state variable.
This indexing is consistently used in Pigreads and NumPy.

The calculation of the diffusion term :math:`\\underline{P} \\nabla \\cdot
\\mathbf D \\nabla \\underline{u}` is implemented as a weighted sum of neighbouring points.
The weights can be calculated using the function :py:func:`weights`, which also requires the
diffusivity_matrix :math:`D` as input, which is set using :py:func:`diffusivity_matrix`::

    diffusivity = pig.diffusivity_matrix(Df=0.03)
    weights = pig.weights(dz, dy, dx, inhom, diffusivity)

Finally, the simulation can be started using :py:func:`run`, which does ``Nt``
forward-Euler steps and only returns the final states after those steps::

    Nt = 200
    dt = 0.025
    for it in range(states.shape[0] - 1):
        states[it + 1] = pig.run(models, inhom, weights, states[it], Nt=Nt, dt=dt)

Full examples can be found in the ``examples`` folder in the Git repository of
this project.
"""

from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from typing import Any

import numpy as np
import yaml

from pigreads import _core, _core_double  # pylint: disable=import-self

from ._core import models
from ._version import version as __version__

dtype: type = np.single
cl: _core.OpenCLWrapper = _core.setup()
cl_double: _core_double.OpenCLWrapper_double = _core_double.setup()


def weights(
    dz: float = 1.0,
    dy: float = 1.0,
    dx: float = 1.0,
    inhom: np.ndarray[Any, Any] | None = None,
    diffusivity: np.ndarray[Any, Any] | float = 1.0,
    double_precision: bool = False,
) -> np.ndarray[Any, Any]:
    """
    Calculate the weights for the diffusion term in the reaction-diffusion
    equation.

    :param dz: The distance between points in the z-dimension, see :py:func:`deltas`.
    :param dy: The distance between points in the y-dimension.
    :param dx: The distance between points in the x-dimension.
    :param inhom: A 3D array with integer values, encoding which model to use at each point. \
            Its value is zero for points outside the medium and one or more for points inside. \
            If ``None``, all points are considered inside the medium.
    :param diffusivity: The diffusivity matrix, see :py:func:`diffusivity_matrix`. \
            If a scalar is given, the matrix is isotropic with the same value in all directions.
    :param double_precision: If ``True``, use double precision for calculations.
    :return: Weight matrix for the diffusion term, A 5D array of shape (1, Nz, Ny, Nx, 19).
    """

    assert dz > 0
    assert dy > 0
    assert dx > 0

    if inhom is None:
        inhom = np.ones(shape=(1, 1, 1))
    assert inhom.ndim == 3

    mask = inhom > 0
    mask.shape = (1, *mask.shape, 1)

    diffusivity = np.array(diffusivity)
    assert isinstance(diffusivity, np.ndarray)
    if diffusivity.ndim == 0:
        diffusivity = diffusivity_matrix(Df=float(diffusivity.item()))
    assert diffusivity.ndim == 4
    assert diffusivity.shape[-1] == 6

    args = dz, dy, dx, mask, diffusivity

    if double_precision:
        return _core_double.weights(cl_double, *args)
    return _core.weights(cl, *args)


def run(
    models: Models,  # pylint: disable=redefined-outer-name
    inhom: np.ndarray[Any, Any],
    weights: np.ndarray[Any, Any],  # pylint: disable=redefined-outer-name
    states: np.ndarray[Any, Any],
    stim_signal: np.ndarray[Any, Any] | None = None,
    stim_shape: np.ndarray[Any, Any] | None = None,
    Nt: int = 1,  # pylint: disable=invalid-name
    dt: float = 0.001,
    double_precision: bool = False,
) -> np.ndarray[Any, Any]:
    """
    Run a Pigreads simulation.

    :param models: The models to be used in the simulation, see :py:class:`Models`.
    :param inhom: A 3D array with integer values, encoding which model to use at each point. \
            Its value is zero for points outside the medium and one or more for points inside. \
            Values larger than zero are used to select one of multiple models: \
            1 for ``models[0]``, 2 for ``models[1]``, etc.
    :param weights: The weights for the diffusion term, see :py:func:`weights`.
    :param states: The initial states of the simulation, a 4D array of shape \
            (Nz, Ny, Nx, Nv), see :py:func:`Models.resting_states`.
    :param stim_signal: A 2D array with the stimulus signal at each time point \
            for all variables, with shape (Nt, Nv). If ``None``, no stimulus is applied.
    :param stim_shape: A 3D array specifying the shape of the stimulus, \
            with shape (Nz, Ny, Nx). If ``None``, no stimulus is applied
    :param Nt: The number of time steps to run the simulation for.
    :param dt: The time step size.
    :param double_precision: If ``True``, use double precision for calculations.
    :return: The final states of the simulation, a 4D array of shape (Nz, Ny, Nx, Nv).
    """
    assert Nt > 0
    assert dt > 0
    assert isinstance(models, Models)
    assert len(models) > 0, "must add at least one model"
    assert inhom.ndim == 3

    if stim_signal is None:
        stim_signal = np.zeros((1, 1, 1, 1, 1))
    else:
        assert stim_signal.ndim == 2
        stim_signal = np.reshape(
            stim_signal, (stim_signal.shape[0], 1, 1, 1, stim_signal.shape[1])
        )

    if stim_shape is None:
        stim_shape = np.zeros((1, 1, 1, 1, 1))
    else:
        assert stim_shape.ndim == 3
        stim_shape = np.where(inhom > 0, stim_shape, 0)
        stim_shape.shape = (1, *stim_shape.shape, 1)  # type: ignore[union-attr]

    args = (
        models.array,
        np.reshape(inhom, (*inhom.shape, 1)),
        weights,
        states,
        stim_signal,
        stim_shape,
        Nt,
        dt,
    )

    if double_precision:
        return _core_double.run(cl_double, *args)  # type: ignore[arg-type]
    return _core.run(cl, *args)  # type: ignore[arg-type]


def get_upper_triangle(
    matrix: np.ndarray[Any, Any],
) -> np.ndarray[Any, Any]:
    """
    Convert a 3x3 matrix to a 6D vector, with the diagonal and upper triangle
    of the matrix as elements in the order xx, yy, zz, yz, xz, xy. Additional
    dimensions are supported, but the last two dimensions must each have size
    3.

    :param matrix: A 3x3 matrix.
    :return: A 6D vector.
    """

    matrix = np.array(matrix)
    assert matrix.ndim >= 2, "matrix must have at least two dimensions"
    assert matrix.shape[-1] == 3, "matrix must be a 3x3 matrix (in the last two axes)"
    assert matrix.shape[-2] == 3, "matrix must be a 3x3 matrix (in the last two axes)"
    assert np.allclose(matrix[..., 1, 2], matrix[..., 2, 1]), (
        "The yz and zy components of matrix need to be the same"
    )
    assert np.allclose(matrix[..., 0, 2], matrix[..., 2, 0]), (
        "The xz and zx components of matrix need to be the same"
    )
    assert np.allclose(matrix[..., 0, 1], matrix[..., 1, 0]), (
        "The xy and yx components of matrix need to be the same"
    )
    triag: np.ndarray[Any, Any] = np.stack(
        (
            matrix[..., 0, 0],  # xx
            matrix[..., 1, 1],  # yy
            matrix[..., 2, 2],  # zz
            matrix[..., 1, 2],  # yz
            matrix[..., 0, 2],  # xz
            matrix[..., 0, 1],  # xy
        ),
        axis=-1,
    )
    return triag


def normalise_vector(
    f: np.ndarray[Any, Any] | list[int] | tuple[int, int, int],
) -> np.ndarray[Any, Any]:
    """
    Normalise a 3D vector to unit length.

    :param f: A 3D vector over space with shape (Nz, Ny, Nx, 3).
    :return: A 5D vector with shape (Nz, Ny, Nx, 3, 1).
    """

    f = np.array(f, dtype=dtype)
    assert isinstance(f, np.ndarray)
    assert f.ndim >= 1, "f must be a 3D vector"
    assert f.ndim <= 4, "too many dimensions for f"
    assert f.shape[-1] == 3, "f must be a 3D vector (in the last axis)"
    while f.ndim < 4:
        f.shape = (1, *f.shape)
    norm = np.linalg.norm(f, axis=-1)
    nonzero = norm > 0
    norm.shape = (*norm.shape, 1)
    f[nonzero] /= norm[nonzero]
    f.shape = (*f.shape, 1)
    assert f.ndim == 5, "f must have 5 dimensions: z, y, x, row, col"
    assert f.shape[-1] == 1, "f must be a 3D column vector"
    assert f.shape[-2] == 3, "f must be a 3D column vector"
    return f


def diffusivity_matrix(
    f: np.ndarray[Any, Any] | list[int] | tuple[int, int, int] | None = None,
    n: np.ndarray[Any, Any] | list[int] | tuple[int, int, int] | None = None,
    Df: np.ndarray[Any, Any] | float = 1.0,  # pylint: disable=invalid-name
    Ds: np.ndarray[Any, Any] | float | None = None,  # pylint: disable=invalid-name
    Dn: np.ndarray[Any, Any] | float | None = None,  # pylint: disable=invalid-name
    dtype: type = dtype,  # pylint: disable=redefined-outer-name
) -> np.ndarray[Any, Any]:
    """
    Define a diffusivity matrix :math:`\\textbf D` for the reaction-diffusion equation.

    If ``f`` and ``n`` are given, the matrix is defined as:

    .. math::

        \\textbf D = \\textbf D_s \\textbf I + (\\textbf D_f - \\textbf D_s)
        \\textbf f \\textbf f^\\mathrm{T} + (\\textbf D_n - \\textbf D_s)
        \\textbf n \\textbf n^\\mathrm{T}

    :param f: The main direction of diffusion, i.e., the fibre direction. \
            3D vector over space with shape (Nz, Ny, Nx, 3). Optional if :math:`D_f=D_s=D_n`.
    :param n: The direction of weakest diffusion, i.e., the direction normal to the fibre sheets. \
            A 3D vector over space with shape (Nz, Ny, Nx, 3). Optional if :math:`D_s=D_n`.
    :param Df: The diffusivity in the direction of the fibres, :math:`\\mathbf{f}`.
    :param Ds: The diffusivity in the fibre sheets, but normal to :math:`\\mathbf{f}`. \
            If ``None``, :math:`D_s` is set to :math:`D_f`.
    :param Dn: The diffusivity in the direction normal to the fibre sheets, \
            i.e., along :math:`\\mathbf{n}`. \
            If ``None``, :math:`D_n` is set to :math:`D_s`.
    :param dtype: The data type of the output array.
    """

    if Ds is None:
        Ds = Df

    if Dn is None:
        Dn = Ds

    if f is None:
        assert np.allclose(Df, Ds), "If Df!=Ds, f must be given"
        f = [0, 0, 0]

    if n is None:
        assert np.allclose(Ds, Dn), "If Ds!=Dn, n must be given"
        n = [0, 0, 0]

    f = normalise_vector(f)
    n = normalise_vector(n)
    eye = np.eye(3, dtype=dtype)
    eye.shape = (1, 1, 1, 3, 3)

    D = get_upper_triangle(  # pylint: disable=invalid-name
        Ds * eye
        + (Df - Ds) * f * np.swapaxes(f, -1, -2)
        + (Dn - Ds) * n * np.swapaxes(n, -1, -2)
    )
    assert D.ndim == 4
    assert D.shape[-1] == 6
    return D


class Models(list[dict[str, Any]]):
    """
    This class stores the models to be used in a Pigreads simulation. It
    behaves like a dictionary :py:class:`dict`. The models are defined in C++
    code. Their metadata is stored YAML format in the :py:func:`models`
    function. The class variable :py:attr:`available` is a dictionary of all
    available models.

    Models are added to an instance of this class using the :py:meth:`add`
    method, or as a list of tuples in the constructor. The order of the models
    is the order in which they are used in the simulation. ``models[0]`` is used
    at ``inhom`` values of 1, ``models[1]`` at 2, etc.

    :param tuples_key_kwargs: A list of tuples with the model key and keyword \
    arguments encoding parameter names and values to be passed to the \
    :py:meth:`add` method. If a string is given, it is treated the name of the \
    first model to be added with no keyword arguments. If ``None``, no \
    models are added.
    """

    available: dict[str, dict[str, Any]] = yaml.safe_load(models())
    """
    Dictionary of all available models.
    """

    def __init__(
        self, tuples_key_kwargs: list[tuple[str, dict[str, float]]] | str | None = None
    ):
        super().__init__(self)
        if tuples_key_kwargs is None:
            tuples_key_kwargs = []
        elif isinstance(tuples_key_kwargs, str):
            tuples_key_kwargs = [(tuples_key_kwargs, {})]
        for key, kwargs in tuples_key_kwargs:
            self.add(key, **kwargs)

    def add(self, key: str, **kwargs: Any) -> None:
        """
        Add a model to the list of models.

        :param key: The key of the model to be added.
        :param kwargs: Parameter names and their values to be passed to the model.
        """
        model = deepcopy(self.info(key))
        for param, value in kwargs.items():
            assert param in model["parameters"], (
                f"Unknown parameter: {param}.\n\n"
                + yaml.safe_dump(dict(model["parameters"]), sort_keys=False)
            )
            model["parameters"][param] = value
        model["key"] = key
        self.append(model)

    @classmethod
    def info(cls, key: str) -> dict[str, Any]:
        """
        Get the metadata of a model.

        :param key: The key of the model.
        :return: A dictionary with the metadata of the model.
        """
        try:
            return deepcopy(cls.available[key])
        except KeyError as e:
            e.add_note(  # type: ignore[attr-defined]
                "\nAvailable models:\n- " + "\n- ".join(cls.available.keys()) + "\n"
            )
            raise e

    @property
    def array(self) -> np.ndarray[Any, Any]:
        """
        Represent the models as a 2D array with the model IDs and all
        parameters to be passed to the C++ code.
        """
        arr = []
        for model in self:
            arr.append(model["id"])
            arr.extend(model["parameters"].values())
        ret: np.ndarray[Any, Any] = np.array(arr, dtype=dtype)
        return ret

    @property
    def Nv(self) -> int:  # pylint: disable=invalid-name
        """
        The maximum number of state variables in the models.
        """
        return max(len(model["variables"]) for model in self)

    def resting_states(
        self,
        inhom: np.ndarray[Any, Any],
        Nframes: int = 1,  # pylint: disable=invalid-name
    ) -> np.ndarray[Any, Any]:
        """
        Create an array of states and fill the first frame with the resting
        values of the models depending on the ``inhom`` values.

        :param inhom: A 3D array with integer values, encoding which model to use at each point. \
                Its value is zero for points outside the medium and one or more for points inside. \
                Values larger than zero are used to select one of multiple models: \
                1 for ``models[0]``, 2 for ``models[1]``, etc.

        :param Nframes: The number of frames in time.
        :return: A 5D array of shape (Nframes, Nz, Ny, Nx, Nv).
        """

        model_count = len(self)
        assert model_count > 0, "must add at least one model"
        assert inhom.ndim == 3
        inhom = inhom.astype(int)
        mask = inhom > 0
        states: np.ndarray[Any, Any] = np.zeros(
            shape=(Nframes, *inhom.shape, self.Nv), dtype=dtype
        )
        for imodel, model in enumerate(self):
            for iv, resting in enumerate(model["variables"].values()):
                states[0, mask * ((inhom - 1) % model_count == imodel), iv] = resting
        return states


def to_ithildin(
    framedur: float,
    dt: float,
    dz: float,
    dy: float,
    dx: float,
    models: Models,  # pylint: disable=redefined-outer-name
    states: np.ndarray[Any, Any],
    inhom: np.ndarray[Any, Any] | None = None,
) -> tuple[dict[str, Any], dict[str, np.ndarray[Any, Any]]]:
    """
    Convert the output of a Pigreads simulation to an Ithildin SimData object.

    While originally designed for a different reaction-diffusion solver,
    the Python module for Ithildin is useful to analyse Pigreads simulations.

    :param framedur: The duration between subsequent frames, usually in milliseconds.
    :param dt: The time step size used in :py:func:`run`, usually in milliseconds.
    :param dz: The distance between points in the z-dimension, see :py:func:`deltas`.
    :param dy: The distance between points in the y-dimension.
    :param dx: The distance between points in the x-dimension.
    :param models: The models used in the simulation, see :py:class:`Models`.
    :param states: The states of the simulation, a 5D array of shape \
            (Nt, Nz, Ny, Nx, Nv), see :py:func:`Models.resting_states` and :py:func:`run`.
    :param inhom: A 3D array with integer values, encoding which model to use at each point. \
            Its value is zero for points outside the medium and one or more for points inside. \
            If ``None``, all points are considered inside the medium.
    :return: Tuple of an Ithildin log file as a string \
            and a dictionary of the variables by variable name.

    Usage::

        import ithildin as ith
        log, variables = pig.to_ithildin(Nt * dt, dt, dz, dy, dx, models, states, inhom)
        sd = ith.SimData(log=ith.Log(data=log))
        sd.vars = variables
    """

    Nt, Nz, Ny, Nx, _ = states.shape  # pylint: disable=invalid-name

    timestamp = datetime.now()

    log = {
        "Ithildin log version": 2,
        "Simulation parameters": {
            "Ithildin library version": f"pigreads {__version__}",
            "Timestep dt": dt,
            "Frame duration": framedur,
            "Number of frames to take": Nt,
            "Serial number": int(timestamp.strftime(r"%Y%m%d%H%M%S")),
            "Name of simulation series": "pigreads",
        },
        "Geometry parameters": {
            "Number of dimensions": 3,
            "Voxel size": [dx, dy, dz],
            "Domain size": [Nx, Ny, Nz],
        },
        "Start date": timestamp.isoformat(),
    }

    for i, model in enumerate(models):
        key = "Model parameters"
        if i > 0:
            key += f" {i}"
        log[key] = {
            "Model type": model["name"],
            "Class": model["key"],
            "Citation": "\n".join(model["dois"]),
            "Parameters": model["parameters"],
            "Initial values": model["variables"],
            "Variable names": list(model["variables"].keys()),
            "Number of vars": len(model["variables"]),
        }

    shape = (-1, Nz, Ny, Nx)
    variables: dict[str, np.ndarray[Any, Any]] = {
        v: states[..., iv].reshape(shape)
        for iv, v in enumerate(models[0]["variables"].keys())
    }
    if inhom is not None:
        variables["inhom"] = inhom.reshape(shape)

    return log, variables


def delta(x: np.ndarray[Any, Any], ax: int = -1) -> float:
    """
    Extract grid spacing from a 3D array.

    :param x: A 3D array.
    :param ax: The axis along which to calculate the distance.
    :return: The distance between the first two points.

    For example, consider this code::

        z, y, x = np.mgrid[0, 0:4:0.2, 0:1:5j]
        dx = pig.delta(z, ax=-1)
        dy = pig.delta(z, ax=-2)
        dz = pig.delta(z, ax=-3)
    """
    assert x.ndim == 3
    diff = np.diff(np.moveaxis(x, ax, -1)[0, 0, :2])
    return 1.0 if diff.shape[0] == 0 else float(diff[0])


def deltas(*x: np.ndarray[Any, Any]) -> list[float]:
    """
    Extract grid spacing from a 3D meshgrid.

    For example, consider this code::

        z, y, x = np.mgrid[0, 0:4:0.2, 0:1:5j]
        dz, dy, dx = pig.deltas(z, y, x)

    :param x: A 3D array.
    :return: A list with the distances between the points.
    """
    return [delta(xi, i) for i, xi in enumerate(x)]


__all__ = [
    "Models",
    "__version__",
    "delta",
    "deltas",
    "diffusivity_matrix",
    "get_upper_triangle",
    "models",
    "normalise_vector",
    "run",
    "to_ithildin",
    "weights",
]



from __future__ import annotations

import numpy as np
import pytest

import pigreads as pig


def test_ndim():
    with pytest.raises(AssertionError):
        pig.get_upper_triangle(np.array([]))


def test_22():
    with pytest.raises(AssertionError):
        pig.get_upper_triangle(np.ones((2, 2)))


def test_32():
    with pytest.raises(AssertionError):
        pig.get_upper_triangle(np.ones((3, 2)))


def test_23():
    with pytest.raises(AssertionError):
        pig.get_upper_triangle(np.ones((2, 3)))


def test_symmetry():
    M = np.arange(9).reshape((3, 3))
    with pytest.raises(AssertionError):
        pig.get_upper_triangle(M)


def test_33():
    M = np.arange(9).reshape((3, 3))
    M[1, 0] = M[0, 1]
    M[2, 0] = M[0, 2]
    M[2, 1] = M[1, 2]
    T = pig.get_upper_triangle(M)
    assert T.ndim == 1
    assert T.shape == (6,)
    assert np.allclose(T, [0, 4, 8, 5, 2, 1])


def test_9933():
    shape = (9, 9, 3, 3)
    M = np.arange(np.prod(shape)).reshape(shape)
    M[..., 1, 0] = M[..., 0, 1]
    M[..., 2, 0] = M[..., 0, 2]
    M[..., 2, 1] = M[..., 1, 2]
    T = pig.get_upper_triangle(M)
    assert T.ndim == 3
    assert T.shape == (*shape[:2], 6)



"""
Type annotations for low-level implementation
---------------------------------------------
"""

from __future__ import annotations

from typing import Any

import numpy as np

# pylint: disable=unused-argument
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

class OpenCLWrapper_double: ...  # pylint: disable=invalid-name

def setup() -> OpenCLWrapper_double: ...
def models() -> str: ...
def weights(
    cl: OpenCLWrapper_double,
    dz: Any,
    dy: Any,
    dx: Any,
    py_mask: np.ndarray[Any, Any],
    py_diffusivity: np.ndarray[Any, Any],
) -> np.ndarray[Any, Any]: ...
def run(
    cl: OpenCLWrapper_double,
    py_models: np.ndarray[Any, Any],
    py_inhom: np.ndarray[Any, Any],
    py_weights: np.ndarray[Any, Any],
    py_states: np.ndarray[Any, Any],
    py_stim_signal: np.ndarray[Any, Any],
    py_stim_shape: np.ndarray[Any, Any],
    Nt: Any,  # pylint: disable=invalid-name
    dt: Any,
) -> np.ndarray[Any, Any]: ...



from __future__ import annotations

import numpy as np
import pytest

import pigreads as pig


def test_no_Ds():
    assert np.allclose(pig.diffusivity_matrix(Df=2), pig.diffusivity_matrix(Df=2, Ds=2))


def test_no_Df():
    assert np.allclose(
        pig.diffusivity_matrix(Df=2, Ds=2), pig.diffusivity_matrix(Df=2, Ds=2, Dn=2)
    )


def test_no_f():
    with pytest.raises(AssertionError):
        pig.diffusivity_matrix(Df=1, Ds=0.1)


def test_f():
    D = pig.diffusivity_matrix(Df=1, Ds=0.5, f=[1, 0, 0])
    assert D.shape == (1, 1, 1, 6)
    assert np.allclose([1, 0.5, 0.5, 0, 0, 0], D)


def test_no_n():
    with pytest.raises(AssertionError):
        pig.diffusivity_matrix(Df=1, Ds=0.5, Dn=0.1, f=[1, 0, 0])


def test_xyz():
    D = pig.diffusivity_matrix(Df=1, Ds=0.5, Dn=0.1, f=[1, 0, 0], n=[0, 1, 0])
    assert D.shape == (1, 1, 1, 6)
    assert np.allclose(D, [1, 0.1, 0.5, 0, 0, 0])


def test_123():
    D = pig.diffusivity_matrix(Df=1, Ds=0.5, Dn=0.1, f=[1, 2, 3], n=[0, 1, 2])
    assert D.shape == (1, 1, 1, 6)
    assert np.allclose(
        D, [0.53571427, 0.56285715, 0.5014285, 0.05428569, 0.10714285, 0.07142857]
    )


def test_multiple():
    random = np.random.default_rng(seed=0)
    a = random.uniform(size=(3,))
    f = np.array([a, a])
    n = np.array([a, a])
    D = pig.diffusivity_matrix(Df=1, Ds=0.5, Dn=0.1, f=f, n=n)
    assert D.shape == (1, 1, 2, 6)
    assert np.allclose(D[0, 0, 0], D[0, 0, 1])



#!/bin/env python3
"""
Run a simulation
"""

from __future__ import annotations

import sys
from pathlib import Path

import ithildin as ith
import numpy as np

import pigreads as pig


def main():
    random = np.random.default_rng(seed=0)

    R = 11.05
    z, y, x = np.mgrid[0:1, -R:R:200j, -R:R:200j]
    dz, dy, dx = pig.deltas(z, y, x)
    r = np.linalg.norm((x, y, z), axis=0)

    inhom = np.ones_like(x, dtype=int)
    inhom[r > R] = 0
    inhom *= random.uniform(0, 1, size=x.shape) > 0.1

    models = pig.Models()
    models.add("marcotte2017dynamical", diffusivity_u=1.0, diffusivity_v=0.05)

    states = models.resting_states(inhom, Nframes=100)
    states[0, np.linalg.norm(((x + 8), y, z), axis=0) < 2, 0] = 1
    states[0, y < 0, 1] = 2

    diffusivity = pig.diffusivity_matrix(Df=0.03)

    weights = pig.weights(dz, dy, dx, inhom, diffusivity)

    Nt = 200
    dt = 0.025
    for it in range(states.shape[0] - 1):
        states[it + 1] = pig.run(models, inhom, weights, states[it], Nt=Nt, dt=dt)

    # plot data with Matplotlib
    import matplotlib.pyplot as plt

    plt.imshow(states[-1, 0, :, :, 0])
    plt.show()

    # write to disk via Ithildin
    log, variables = pig.to_ithildin(Nt * dt, dt, dz, dy, dx, models, states, inhom)
    sd = ith.SimData(log=ith.Log(data=log))
    sd.vars = variables
    sd.to_npz("results/" + Path(sys.argv[0]).stem + ".npz")


if __name__ == "__main__":
    main()



#!/bin/env python3
"""
Convert a full model YAML file to a C++ header file for OpenCL to use in
Pigreads.
"""

from __future__ import annotations

import re
import sys
from pathlib import Path
from textwrap import dedent, indent

import yaml


def str_presenter(dumper, data):
    """
    A YAML string presenter that uses block style for multiline strings.
    """
    if len(data.splitlines()) > 1:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.representer.SafeRepresenter.add_representer(str, str_presenter)

if __name__ == "__main__":
    path = Path(sys.argv[1])
    with path.open("r", encoding="utf-8") as f:
        model = yaml.safe_load(f)

    key = model["key"] = re.sub(r"\.(model|float|double)$", "", path.stem)

    source = model["source"]
    del model["source"]

    if "diffusivity" not in model:
        model["diffusivity"] = {}
    if "parameters" not in model:
        model["parameters"] = {}

    diffusivity_by_varname = model["diffusivity"]
    diffusivity = {
        "diffusivity_" + varname: value
        for varname, value in diffusivity_by_varname.items()
    }
    model["parameters"] = {**diffusivity, **model["parameters"]}
    del model["diffusivity"]

    model["description"] = dedent(model.get("description", "")).strip()

    with Path(sys.argv[2]).open("w", encoding="utf-8") as f:
        f.write(
            dedent(rf"""
            #ifndef __OPENCL_VERSION__
            const std::string Model_{key}_info = R"model(
        """).strip()
            + "\n"
        )
        f.write(
            yaml.safe_dump(
                {key: model}, sort_keys=False, indent=2, allow_unicode=True
            ).strip()
            + "\n"
        )
        f.write(
            dedent(rf"""
            )model";
            #endif

            static const Size Model_{key}_id = UNIQUE_ID;
            static const Size Model_{key}_Nv = {len(model["variables"])};
            static const Size Model_{key}_Np = {len(model["parameters"])};

            #ifdef __OPENCL_VERSION__
            void Model_{key}_step(
                    Real* const params,
                    struct States weights,
                    struct States states_old,
                    struct States states_new,
                    const Real dt
            ) {{
        """).strip()
            + "\n"
        )
        for ip, param in enumerate(model["parameters"]):
            f.write(f"  const Real {param} = params[{ip}];\n")
        for ivar, varname in enumerate(model["variables"]):
            f.write(f"  const Real {varname} = _r(_v({ivar}, states_old));\n")
            f.write(f"  Real* const _new_{varname} = _pr(_v({ivar}, states_new));\n")
            if varname in diffusivity_by_varname:
                f.write(
                    f"  const Real _diffuse_{varname} = diffusivity_{varname} * diffuse(weights, _v({ivar}, states_old));\n"
                )
        f.write("\n" + indent(source.strip(), "  ") + "\n")
        f.write(
            dedent(rf"""
            }}
            #endif

            #ifdef __OPENCL_VERSION__
            __kernel void Model_{key}_kernel(
                    Size model_count,
                    __global Size* model_ids,
                    __global Size* model_offsets,
                    __global Real* model_params,
                    __global void* inhom_data,      struct StatesIdx inhom_idx,
                    __global void* weights_data,    struct StatesIdx weights_idx,
                    __global void* states_old_data, struct StatesIdx states_old_idx,
                    __global void* states_new_data, struct StatesIdx states_new_idx,
                    const Real dt
            ) {{

              struct States inhom      = {{inhom_data,      STATES_UNPACK(inhom_idx)}};
              struct States weights    = {{weights_data,    STATES_UNPACK(weights_idx)}};
              struct States states_old = {{states_old_data, STATES_UNPACK(states_old_idx)}};
              struct States states_new = {{states_new_data, STATES_UNPACK(states_new_idx)}};

              const Size iz = get_global_id(0);
              const Size iy = get_global_id(1);
              const Size ix = get_global_id(2);

              if (ix < states_old.Nx && iy < states_old.Ny && iz < states_old.Nz) {{
                const Int inhom_zyx = _i(States_offset(inhom, 0, iz, iy, ix, 0));
                if (inhom_zyx > 0) {{
                  const Size imodel = (inhom_zyx - 1) % model_count;
                  const Size model_id = model_ids[imodel];
                  if (model_id == Model_{key}_id) {{
                    Real* params = model_params + model_offsets[imodel];
                    struct States w = States_offset(weights, 0, iz, iy, ix, 0);
                    struct States u = States_offset(states_old, 0, iz, iy, ix, 0);
                    struct States u_ = States_offset(states_new, 0, iz, iy, ix, 0);
                    Model_{key}_step(params, w, u, u_, dt);
                  }}
                }}
              }}
            }}
            #endif
        """).strip()
            + "\n"
        )



"""
Type annotations for low-level implementation
---------------------------------------------
"""

from __future__ import annotations

from typing import Any

import numpy as np

# pylint: disable=unused-argument
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

class OpenCLWrapper: ...

def setup() -> OpenCLWrapper: ...
def models() -> str: ...

# def models_double() -> str: ...

def weights(
    cl: OpenCLWrapper,
    dz: Any,
    dy: Any,
    dx: Any,
    py_mask: np.ndarray[Any, Any],
    py_diffusivity: np.ndarray[Any, Any],
) -> np.ndarray[Any, Any]: ...

# def weights_double(
#     cl: OpenCLWrapper,
#     dz: Any,
#     dy: Any,
#     dx: Any,
#     py_mask: np.ndarray[Any, Any],
#     py_diffusivity: np.ndarray[Any, Any],
# ) -> np.ndarray[Any, Any]: ...

def run(
    cl: OpenCLWrapper,
    py_models: np.ndarray[Any, Any],
    py_inhom: np.ndarray[Any, Any],
    py_weights: np.ndarray[Any, Any],
    py_states: np.ndarray[Any, Any],
    py_stim_signal: np.ndarray[Any, Any],
    py_stim_shape: np.ndarray[Any, Any],
    Nt: Any,  # pylint: disable=invalid-name
    dt: Any,
) -> np.ndarray[Any, Any]: ...

# def run_double(
#     cl: OpenCLWrapper,
#     py_models: np.ndarray[Any, Any],
#     py_inhom: np.ndarray[Any, Any],
#     py_weights: np.ndarray[Any, Any],
#     py_states: np.ndarray[Any, Any],
#     py_stim_signal: np.ndarray[Any, Any],
#     py_stim_shape: np.ndarray[Any, Any],
#     Nt: Any,  # pylint: disable=invalid-name
#     dt: Any,
# ) -> np.ndarray[Any, Any]: ...



#!/bin/env python3
"""
Export Myokit MMT files to Pigreads C++ code
--------------------------------------------
"""

from __future__ import annotations

from typing import Any

import myokit as mk  # type: ignore[import-not-found]
import yaml
from myokit.formats.opencl import (  # type: ignore[import-not-found]
    OpenCLExpressionWriter,
)


def str_presenter(dumper, data):
    """
    A YAML string presenter that uses block style for multiline strings.
    """
    if len(data.splitlines()) > 1:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.representer.SafeRepresenter.add_representer(str, str_presenter)


class PigreadsMyokitWriter:
    """
    This class implements the conversion of a Myokit model to Pigreads
    compatible C++ code.

    :param model: The Myokit model to convert.
    :param meta: A dictionary with metadata to include in the generated code.
    :param double_precision: If ``True``, use double precision for calculations.

    :var model: The Myokit model to convert.
    :vartype model: myokit.Model
    :var exwr: The C++ expression writer.
    :vartype exwr: myokit.formats.opencl.OpenCLExpressionWriter
    :var states: The list of states in the model.
    :vartype states: list[mk.State]
    :var meta: A dictionary with metadata to include in the generated code.
    :vartype meta: dict[str, Any]
    :var double_precision: If ``True``, use double precision for calculations.
    :vartype double_precision: bool
    """

    def __init__(
        self, model: mk.Model, meta: dict[str, Any], double_precision: bool = False
    ):  # pylint: disable=redefined-outer-name
        self.model = model
        if double_precision:
            self.exwr = OpenCLExpressionWriter(
                precision=mk.DOUBLE_PRECISION, native_math=False
            )
        else:
            self.exwr = OpenCLExpressionWriter(
                precision=mk.SINGLE_PRECISION, native_math=True
            )
        self.exwr.set_lhs_function(self.lhs_format)
        self.states = list(model.states())
        self.meta = meta
        self.generate_variable_abbreviations()

    @property
    def diffusivities(self) -> dict[str, float]:
        """
        The diffusivities of the model as defined in the metadata.
        """
        d = self.meta.get("diffusivity", {})
        assert isinstance(d, dict)
        return d

    def get_ivar(self, varname: str) -> int:
        """
        Get the index of a state variable by its name.

        :param varname: The name of the state variable.
        :return: The index of the state variable.
        """
        return next(
            i
            for i, state in enumerate(self.states)
            if str(self.lhs_format(state)) == varname
        )

    @staticmethod
    def nodots(s: Any) -> str:
        return str(s).replace(".", "_")

    def lhs_format(self, x: mk.LhsExpression):
        """
        Format a left-hand side expression.

        :param x: The left-hand side expression to format.
        :return: The formatted left-hand side expression.
        """
        assert not isinstance(x, mk.Derivative), "Can not handle derivatives here."
        if isinstance(x, mk.Name):
            return self.lhs_format(x.var())
        s = self.nodots(x)
        return self.variable_abbreviations.get(s, s)

    def state_equation(self, q: mk.Equation):
        """
        Format a state equation.

        :param q: The state equation to format.
        :return: The formatted state equation.
        """
        w = q.lhs.walk()
        next(w)
        v = next(w)
        vin = str(self.lhs_format(v))
        rhs = str(self.exwr.ex(q.rhs))
        if vin in self.diffusivities:
            rhs += f" + _diffuse_{vin}"
        return f"*_new_{vin} = {vin} + dt*({rhs});"

    def generate_variable_abbreviations(self) -> None:
        variables = dict(
            zip(self.model.states(), self.model.initial_values(as_floats=True))
        )
        variables_long = {self.nodots(v): f for v, f in variables.items()}
        variables_short = {}
        for variable, value in variables.items():
            short_varname = str(variable).rsplit(".", maxsplit=1)[-1]
            if short_varname not in variables_short:
                variables_short[short_varname] = value
            else:
                variables_short = variables_long
                break
        self.variables = variables_short
        self.variable_abbreviations = dict(
            zip(variables_long.keys(), variables_short.keys())
        )

    def __str__(self) -> str:
        """
        Convert the model to a string. This is the main entry point for the
        conversion.

        :return: The model as a string.
        """
        parameters = {
            self.lhs_format(q.lhs): float(q.rhs)
            for block in self.model.solvable_order().values()
            for q in block
            if q.rhs.is_constant()
        }

        meta = {  # pylint: disable=redefined-outer-name
            "name": rf"{self.model.name()} (exported from Myokit)",
            "description": str(self.model.meta.get("desc", "")),
            "dois": [],
            "variables": self.variables,
            "parameters": parameters,
            **self.meta,
        }

        step = []
        for blockname, block in self.model.solvable_order().items():
            block_ = [eq for eq in block if not eq.rhs.is_constant()]
            if len(block_) < 1:
                continue
            step.append("")
            step.append(f"// {blockname}")
            for q in block_:
                if q.lhs.is_derivative():
                    step.append(self.state_equation(q))
                else:
                    step.append("const Real " + self.exwr.eq(q) + ";")

        meta["source"] = "\n".join(step).strip()
        return yaml.safe_dump(meta, sort_keys=False, indent=2, allow_unicode=True)


if __name__ == "__main__":
    import sys
    from pathlib import Path

    model, protocol, script = mk.load(sys.argv[1])
    with Path(sys.argv[2]).open("r", encoding="utf-8") as f:
        meta = yaml.safe_load(f)
    writer = PigreadsMyokitWriter(model, meta, double_precision=sys.argv[3] == "double")
    with Path(sys.argv[4]).open("w", encoding="utf-8") as f:
        f.write(str(writer))



from __future__ import annotations

from typing import Any

import numpy as np

import pigreads as pig


def test_inhom():
    z, y, x = np.mgrid[0:1, 0:1, 0:3]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.arange(3).reshape((1, 1, 3))
    models = pig.Models()
    models.add("aliev1996simple")
    models.add("gray1983autocatalytic")
    states = models.resting_states(inhom, Nframes=2)
    states[0, 0, 0, 0, :] = 1.0  # outside
    states[0, 0, 0, 1, 0] = 0.4  # aliev1996simple u
    states[0, 0, 0, 2, 1] = 0.5  # gray1983autocatalytic v
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=0)
    states[1] = pig.run(models, inhom, weights, states[0], Nt=2, dt=0.1)
    assert np.allclose(states[0, 0, 0, 0, :], 1), "starting at given outside value"
    assert np.allclose(states[1, 0, 0, 0, :], 0), "outside value must be ignored"
    assert states[0, 0, 0, 1, 0] < states[1, 0, 0, 1, 0], (
        "aliev1996simple u must increase"
    )
    assert states[0, 0, 0, 1, 1] < states[1, 0, 0, 1, 1], (
        "aliev1996simple v must increase"
    )
    assert states[0, 0, 0, 2, 0] > states[1, 0, 0, 2, 0], (
        "gray1983autocatalytic u must decrease"
    )
    assert states[0, 0, 0, 2, 1] < states[1, 0, 0, 2, 1], (
        "gray1983autocatalytic v must increase"
    )


def test_stim():
    z, y, x = np.mgrid[0:1, 0:1, 0:5]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    inhom[:, :, -1] = 2
    models = pig.Models()
    models.add("trivial")
    models.add("aliev1996simple", k=0, eps0=0, mu1=0)
    states = models.resting_states(inhom, Nframes=2)
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=0)
    stim_signal = np.array([(0, 0), (-1, 2), (4, 1), (np.nan, np.nan)])
    stim_shape = np.array([0, 1, 0.5, -1, 1]).reshape(inhom.shape)
    dt = 1.0
    states[1] = pig.run(
        models,
        inhom,
        weights,
        states[0],
        stim_signal=stim_signal,
        stim_shape=stim_shape,
        Nt=4,
        dt=dt,
    )
    assert states.shape == (2, 1, 1, 5, 2)
    sum_u, sum_v = np.sum(stim_signal[:3], axis=0)
    for i in range(4):
        assert states[1, 0, 0, i, 0] == dt * stim_shape[0, 0, i] * sum_u, f"i={i}"
    assert states[1, 0, 0, 4, 1] == dt * stim_shape[0, 0, 4] * sum_v


def test_stim_dt():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("trivial")
    states = models.resting_states(inhom, Nframes=3)
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=0)
    stim_signal = np.arange(5).reshape((-1, 1))
    stim_shape = inhom
    dt = 0.1
    states[1] = dt * pig.run(
        models,
        inhom,
        weights,
        states[0],
        stim_signal=stim_signal,
        stim_shape=stim_shape,
        Nt=11,
        dt=1,
    )
    states[2] = pig.run(
        models,
        inhom,
        weights,
        states[0],
        stim_signal=stim_signal,
        stim_shape=stim_shape,
        Nt=11,
        dt=dt,
    )
    assert np.allclose(states[1], states[2])


def test_cell():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("aliev1996simple")
    states = models.resting_states(inhom, Nframes=3)
    states[0, ..., 0] = 0.4
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=0)
    for i in range(states.shape[0] - 1):
        states[i + 1] = pig.run(models, inhom, weights, states[i], Nt=200, dt=0.1)
    assert states.shape == (3, 1, 1, 1, 2)
    states = states.squeeze()
    assert np.allclose(
        states,
        [
            [4.00000006e-01, 0.00000000e00],
            [9.06229615e-01, 5.90469241e-01],
            [4.87683921e-11, 1.08949475e-01],
        ],
    )


def test_cable():
    z, y, x = np.mgrid[0:1, 0:1, 0:50]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    inhom[..., -1] = 0
    models = pig.Models("aliev1996simple")
    states = models.resting_states(inhom, Nframes=3)
    states[0, x < 5, 0] = 0.4
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=1)
    Nt, dt = 400, 0.05
    for i in range(states.shape[0] - 1):
        states[i + 1] = pig.run(models, inhom, weights, states[i], Nt=Nt, dt=dt)
    assert states.shape == (3, 1, 1, 50, 2)
    assert np.allclose(
        states[:, 0, 0, (0, 25, 48), 0],
        [
            [4.0000001e-01, 0.0000000e00, 0.0000000e00],
            [9.0334904e-01, 9.9558353e-01, 1.0182223e-14],
            [1.0085634e-08, 5.3530955e-01, 9.9503601e-01],
        ],
    )


def test_spiral():
    z, y, x = np.mgrid[0:1, -25:25, -25:25]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    inhom[:, -1, :] = 0
    inhom[:, :, -1] = 0
    models = pig.Models("aliev1996simple")
    states = models.resting_states(inhom, Nframes=3)
    states[0, x > 0, 0] = 1
    states[0, y > 0, 1] = 1
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=1)
    Nt, dt = 400, 0.05
    for i in range(states.shape[0] - 1):
        states[i + 1] = pig.run(models, inhom, weights, states[i], Nt=Nt, dt=dt)
    indices = 0, 10, 40, 48, 49
    assert np.allclose(
        states[:, 0, indices, indices, 0],
        [
            [0.0000000e00, 0.0000000e00, 1.0000000e00, 1.0000000e00, 1.0000000e00],
            [2.1328260e-01, 9.8938406e-01, 6.8952787e-08, 1.8248732e-11, 0.0000000e00],
            [8.7120086e-01, 5.8410158e-03, 2.0007190e-05, 4.3070755e-12, 0.0000000e00],
        ],
    )


def test_aniso():
    z, y, x = np.mgrid[0:1, -25:25, -25:25]
    r = np.linalg.norm((x, y), axis=0)
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("aliev1996simple")
    states = models.resting_states(inhom, Nframes=3)
    states[0, r < 3, 0] = 1
    diffusivity = pig.diffusivity_matrix(Df=1.0, Ds=0.2, f=[1, 2, 3])
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=diffusivity)
    Nt, dt = 400, 0.05
    for i in range(states.shape[0] - 1):
        states[i + 1] = pig.run(models, inhom, weights, states[i], Nt=Nt, dt=dt)
    assert np.allclose(states[1, 0, 10, 15, 0], 0.15971428)
    assert np.allclose(states[1, 0, 15, 34, 0], 0.3017956)
    assert np.allclose(states[2, 0, 37, 33, 0], 0.1480847)
    assert np.allclose(states[2, 0, 25, 19, 0], 0.00012683906)


def test_double():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("aliev1996simple")
    states: np.ndarray[Any, Any] = models.resting_states(inhom).astype(np.float64)
    states[0, ..., 0] = 0.4
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=0)
    assert states[0].ndim == 4
    states_single = pig.run(models, inhom, weights, states[0], Nt=200, dt=0.1)
    states_double = pig.run(
        models, inhom, weights, states[0], Nt=200, dt=0.1, double_precision=True
    )
    assert states_single.ndim == 4
    assert states_double.ndim == 4
    assert states_single.dtype == np.float32
    assert states_double.dtype == np.float64
    assert np.allclose(states_single, states_double)


def test_dt():
    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x)
    models = pig.Models("aliev1996simple")
    states = models.resting_states(inhom, Nframes=3)
    states[0, ..., 0] = 0.4
    weights = pig.weights(dz, dy, dx, inhom, diffusivity=0)
    dt = 0.0001
    states[1] = pig.run(models, inhom, weights, states[0], Nt=101, dt=dt)
    states[2] = pig.run(models, inhom, weights, states[0], Nt=11, dt=dt * 10)
    assert np.allclose(states[1], states[2], rtol=1e-3, atol=1e-5)
    assert np.allclose(
        states.squeeze(),
        [
            [4.0000001e-01, 0.0000000e00],
            [4.0485540e-01, 4.8295802e-05],
            [4.0485030e-01, 4.8268699e-05],
        ],
    )



from __future__ import annotations

import argparse
import shutil
from pathlib import Path

import nox

DIR = Path(__file__).parent.resolve()

nox.needs_version = ">=2024.3.2"
nox.options.sessions = ["lint", "pylint", "tests"]
nox.options.default_venv_backend = "uv|virtualenv"


@nox.session
def lint(session: nox.Session) -> None:
    """
    Run the linter.
    """
    session.install("pre-commit")
    session.run(
        "pre-commit", "run", "--all-files", "--show-diff-on-failure", *session.posargs
    )


@nox.session
def pylint(session: nox.Session) -> None:
    """
    Run PyLint.
    """
    # This needs to be installed into the package environment, and is slower
    # than a pre-commit check
    session.install(".", "pylint>=3.2")
    session.run("pylint", "pigreads", *session.posargs)


@nox.session
def tests(session: nox.Session) -> None:
    """
    Run the unit and regular tests.
    """
    session.install(".[test]")
    session.run("pytest", *session.posargs)


@nox.session(reuse_venv=True)
def docs(session: nox.Session) -> None:
    """
    Build the docs. Pass --non-interactive to avoid serving. First positional argument is the target directory.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-b", dest="builder", default="html", help="Build target (default: html)"
    )
    parser.add_argument("output", nargs="?", help="Output directory")
    args, posargs = parser.parse_known_args(session.posargs)
    serve = args.builder == "html" and session.interactive

    session.install("-e.[docs]", "sphinx-autobuild")

    shared_args = (
        "-n",  # nitpicky mode
        "-T",  # full tracebacks
        f"-b={args.builder}",
        "docs",
        args.output or f"docs/_build/{args.builder}",
        *posargs,
    )

    if serve:
        session.run("sphinx-autobuild", "--open-browser", *shared_args)
    else:
        session.run("sphinx-build", "--keep-going", *shared_args)


@nox.session
def build_api_docs(session: nox.Session) -> None:
    """
    Build (regenerate) API docs.
    """

    session.install("sphinx")
    session.run(
        "sphinx-apidoc",
        "-o",
        "docs/api/",
        "--module-first",
        "--no-toc",
        "--force",
        "src/pigreads",
    )


@nox.session
def build(session: nox.Session) -> None:
    """
    Build an SDist and wheel.
    """

    build_path = DIR.joinpath("build")
    if build_path.exists():
        shutil.rmtree(build_path)

    session.install("build")
    session.run("python", "-m", "build")



from __future__ import annotations

import numpy as np

import pigreads as pig


def test_trivial():
    weights = pig.weights()
    assert weights.shape == (1, 1, 1, 1, 19)
    weights = weights.flatten()
    assert weights[0] == -6
    for i in range(1, 7):
        assert weights[i] == 1
    for i in range(7, 19):
        assert weights[i] == 0


def test_scaling():
    assert np.allclose(pig.weights(diffusivity=2), 2 * pig.weights())


def test_matrix():
    diffusivity = pig.diffusivity_matrix(Df=1, Ds=0.5, Dn=0.1, f=[1, 2, 3], n=[4, 5, 6])
    weights = pig.weights(diffusivity=diffusivity)
    assert np.allclose(
        weights.flatten(),
        [
            -3.2,
            0.45259738,
            0.45259738,
            0.512987,
            0.512987,
            0.63441557,
            0.63441557,
            0.02922077,
            -0.02922077,
            -0.02922077,
            0.02922077,
            -0.00876624,
            0.00876624,
            0.00876624,
            -0.00876624,
            -0.01623377,
            0.01623377,
            0.01623377,
            -0.01623377,
        ],
    )


def test_dot():
    z, y, x = np.mgrid[0:5, 0:5, 0:5]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.zeros_like(x, dtype=int)
    inhom[2, 2, 2] = 1
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 5, 5, 5, 19)
    assert np.allclose(weights, 0)


def test_cable_x():
    z, y, x = np.mgrid[0:5, 0:5, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.zeros_like(x, dtype=int)
    inhom[2, 2, :] = 1
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 5, 5, 1, 19)
    weights = weights[0, 2, 2, 0]
    assert weights[0] == -2
    assert weights[1] == 1
    assert weights[2] == 1
    for i in range(3, 19):
        assert weights[i] == 0


def test_cable_y():
    z, y, x = np.mgrid[0:5, 0:1, 0:5]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.zeros_like(x, dtype=int)
    inhom[2, :, 2] = 1
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 5, 1, 5, 19)
    weights = weights[0, 2, 0, 2]
    assert weights[0] == -2
    assert weights[3] == 1
    assert weights[4] == 1
    for i in range(1, 3):
        assert weights[i] == 0
    for i in range(5, 19):
        assert weights[i] == 0


def test_cable_z():
    z, y, x = np.mgrid[0:1, 0:5, 0:5]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.zeros_like(x, dtype=int)
    inhom[:, 2, 2] = 1
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 1, 5, 5, 19)
    weights = weights[0, 0, 2, 2]
    assert weights[0] == -2
    assert weights[5] == 1
    assert weights[6] == 1
    for i in range(1, 5):
        assert weights[i] == 0
    for i in range(7, 19):
        assert weights[i] == 0


def test_edge_x():
    z, y, x = np.mgrid[0:1, 0:1, 0:4]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)
    inhom[:, :, 0] = 0
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 1, 1, 4, 19)
    weights = weights[0, 0, 0, :, :]
    assert np.allclose(np.sum(weights, axis=-1), 0)
    assert np.allclose(weights[0], 0)
    assert np.allclose(weights[2], pig.weights().flatten())
    weights[3, 1], weights[3, 2] = weights[3, 2], weights[3, 1]
    assert np.allclose(weights[1], weights[3])
    assert weights[1, 0] == -5
    assert np.allclose(weights[1][8:], 0)


def test_edge_y():
    z, y, x = np.mgrid[0:1, 0:4, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)
    inhom[:, 0, :] = 0
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 1, 4, 1, 19)
    weights = weights[0, 0, :, 0, :]
    assert np.allclose(np.sum(weights, axis=-1), 0)
    assert np.allclose(weights[0], 0)
    assert np.allclose(weights[2], pig.weights().flatten())
    weights[3, 3], weights[3, 4] = weights[3, 4], weights[3, 3]
    assert np.allclose(weights[1], weights[3])
    assert weights[1, 0] == -5
    assert np.allclose(weights[1][8:], 0)


def test_edge_z():
    z, y, x = np.mgrid[0:4, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)
    inhom[0, :, :] = 0
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 4, 1, 1, 19)
    weights = weights[0, :, 0, 0, :]
    assert np.allclose(np.sum(weights, axis=-1), 0)
    assert np.allclose(weights[0], 0)
    assert np.allclose(weights[2], pig.weights().flatten())
    weights[3, 5], weights[3, 6] = weights[3, 6], weights[3, 5]
    assert np.allclose(weights[1], weights[3])
    assert weights[1, 0] == -5
    assert np.allclose(weights[1][8:], 0)


def test_corner_xy():
    z, y, x = np.mgrid[0:1, 0:4, 0:4]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)
    inhom[:, :, 0] = 0
    inhom[:, 0, :] = 0
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 1, 4, 4, 19)
    weights = weights[0, 0, :, :, :]
    assert np.allclose(np.sum(weights, axis=-1), 0)
    assert np.allclose(weights[0, :], 0)
    assert np.allclose(weights[:, 0], 0)
    assert np.allclose(weights[2, 2], pig.weights().flatten())
    assert weights[+1, +1, 0] == -4
    assert weights[-1, -1, 0] == -4
    assert weights[+1, -1, 0] == -4
    assert weights[-1, -1, 0] == -4
    assert weights[2, -1, 0] == -5
    assert weights[2, +1, 0] == -5
    assert weights[-1, 2, 0] == -5
    assert weights[+1, 2, 0] == -5


def test_corner_xz():
    z, y, x = np.mgrid[0:4, 0:1, 0:4]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)
    inhom[:, :, 0] = 0
    inhom[0, :, :] = 0
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 4, 1, 4, 19)
    weights = weights[0, :, 0, :, :]
    assert np.allclose(np.sum(weights, axis=-1), 0)
    assert np.allclose(weights[0, :], 0)
    assert np.allclose(weights[:, 0], 0)
    assert np.allclose(weights[2, 2], pig.weights().flatten())
    assert weights[+1, +1, 0] == -4
    assert weights[-1, -1, 0] == -4
    assert weights[+1, -1, 0] == -4
    assert weights[-1, -1, 0] == -4
    assert weights[2, -1, 0] == -5
    assert weights[2, +1, 0] == -5
    assert weights[-1, 2, 0] == -5
    assert weights[+1, 2, 0] == -5


def test_cube():
    z, y, x = np.mgrid[0:4, 0:4, 0:4]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)
    inhom[:, :, 0] = 0
    inhom[:, 0, :] = 0
    inhom[0, :, :] = 0
    weights = pig.weights(dz, dy, dx, inhom)
    assert weights.shape == (1, 4, 4, 4, 19)
    weights = weights[0, :, :, :, :]
    assert np.allclose(np.sum(weights, axis=-1), 0)
    assert np.allclose(weights[0, :, :], 0)
    assert np.allclose(weights[:, 0, :], 0)
    assert np.allclose(weights[:, :, 0], 0)
    assert np.allclose(weights[2, 2, 2], pig.weights().flatten())
    assert weights[+1, +1, +1, 0] == -3
    assert weights[+1, +1, -1, 0] == -3
    assert weights[+1, -1, +1, 0] == -3
    assert weights[+1, -1, -1, 0] == -3
    assert weights[-1, +1, +1, 0] == -3
    assert weights[-1, +1, -1, 0] == -3
    assert weights[-1, -1, +1, 0] == -3
    assert weights[-1, -1, -1, 0] == -3
    assert weights[2, +1, +1, 0] == -4
    assert weights[2, -1, -1, 0] == -4
    assert weights[2, +1, -1, 0] == -4
    assert weights[2, -1, -1, 0] == -4
    assert weights[+1, 2, +1, 0] == -4
    assert weights[-1, 2, -1, 0] == -4
    assert weights[+1, 2, -1, 0] == -4
    assert weights[-1, 2, -1, 0] == -4
    assert weights[+1, +1, 2, 0] == -4
    assert weights[-1, -1, 2, 0] == -4
    assert weights[+1, -1, 2, 0] == -4
    assert weights[-1, -1, 2, 0] == -4
    assert weights[2, 2, -1, 0] == -5
    assert weights[2, 2, +1, 0] == -5
    assert weights[2, -1, 2, 0] == -5
    assert weights[2, +1, 2, 0] == -5
    assert weights[-1, 2, 2, 0] == -5
    assert weights[+1, 2, 2, 0] == -5


def test_double():
    z, y, x = np.mgrid[0:4, 0:4, 0:4]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)
    inhom[:, :, 0] = 0
    inhom[:, 0, :] = 0
    inhom[0, :, :] = 0
    single = pig.weights(dz, dy, dx, inhom, double_precision=False)
    double = pig.weights(dz, dy, dx, inhom, double_precision=True)
    assert np.allclose(single, double)



from __future__ import annotations

import re
from typing import Any

import numpy as np
import pytest

import pigreads as pig


def test_structure():
    "Check that the model metadata structured correctly."
    assert isinstance(pig.models(), str)

    models = pig.Models()

    assert isinstance(models.available, dict)
    for name, model in models.available.items():
        assert isinstance(name, str)
        assert isinstance(model, dict)
        for key in model:
            assert isinstance(key, str)
        assert isinstance(model["id"], int)
        assert isinstance(model["name"], str)
        assert isinstance(model["description"], str)
        assert isinstance(model["dois"], list)
        for doi in model["dois"]:
            assert isinstance(doi, str)
            assert re.match(r"^https://doi.org/10\.", doi)
        for dictname in ["variables", "parameters"]:
            assert isinstance(model[dictname], dict)
            for varname, value in model[dictname].items():
                assert isinstance(varname, str)
                assert isinstance(value, (float, int))


def test_data_double():
    assert pig._core.models() == pig._core_double.models()


def test_str():
    pig.Models("marcotte2017dynamical")


def test_info():
    models = pig.Models()
    key = "marcotte2017dynamical"
    assert models.available[key] == models.info(key)


def test_info_empty():
    models = pig.Models()
    with pytest.raises(KeyError):
        models.info("")


def test_add():
    key = "marcotte2017dynamical"
    models = pig.Models(
        [
            (key, {}),
            (key, {"eps": 1}),
        ]
    )
    models.add(key)
    models.add(key, eps=2)
    assert models[0] == models.info(key)
    assert models[0] == models[2]
    assert models[1]["parameters"]["eps"] == 1
    assert models[3]["parameters"]["eps"] == 2


def test_Nv():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    assert models.Nv == 2


def test_resting_states():
    models = pig.Models()
    models.add("gray1983autocatalytic")
    models.add("courtemanche1998ionic")
    inhom = np.array([[[0, 1], [2, 3]]], dtype=int)
    states = models.resting_states(inhom, Nframes=2)
    assert states.shape == (2, *inhom.shape, models.Nv)
    assert states[0, 0, 0, 0, 0] == 0
    assert states[0, 0, 0, 0, 1] == 0
    assert states[0, 0, 0, 1, 0] == 1
    assert states[0, 0, 0, 1, 1] == 0
    assert abs(-81.18 - states[0, 0, 1, 0, 0]) < 1e-6
    assert abs(0.002908 - states[0, 0, 1, 0, 1]) < 1e-6
    assert states[0, 0, 1, 1, 0] == 1
    assert states[0, 0, 1, 1, 1] == 0
    assert np.allclose(states[1], 0)


def test_array():
    models = pig.Models()
    models.add("marcotte2017dynamical", eps=2)
    models.add("gray1983autocatalytic")
    a = models.array
    assert a.ndim == 1
    assert a.shape[0] == 11
    assert a[0] == models[0]["id"]
    assert a[6] == models[1]["id"]
    assert np.allclose(a[1:6], [4.0062, 0.20031, 1.389, 2.0, 1.5415])
    assert np.allclose(a[7:], [1.0, 0.5, 0.055, 0.062])


def test_sim0d():
    z, y, x = np.mgrid[0:1, 0:3, 0:3]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)

    for modelname in pig.Models().available:
        models = pig.Models(modelname)
        states: np.ndarray[Any, Any] = models.resting_states(inhom, Nframes=2).astype(
            np.float32
        )
        weights = pig.weights(dz, dy, dx, inhom, diffusivity=0)
        states[1] = pig.run(models, inhom, weights, states[0], Nt=2, dt=1e-10)
